/*
 * TemplateBinding.java
 */ 

package net.sf.tacos.binding;

import org.apache.commons.lang.StringUtils;
import org.apache.hivemind.Location;
import org.apache.tapestry.IBinding;
import org.apache.tapestry.IComponent;
import org.apache.tapestry.binding.AbstractBinding;
import org.apache.tapestry.binding.BindingFactory;
import org.apache.tapestry.coerce.ValueConverter;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * An implementation of Tapestry {@link IBinding} that provides a binding
 * similar to {@link org.apache.tapestry.binding.LiteralBinding} with the 
 * additional benefit that it can also include variables.
 *
 * The variables should be put into <code>${}</code> .<br/>
 *
 * Here's an example:
 * <code>
 * &lt;binding name="effects" value="template:some id is ${currNote.id}."/&gt;
 * </code>
 * <p/>
 * <p/>
 * The binding can also get its content from resource bundles. It will still
 * go on and process it - looking for variables to substitute.<br/>
 * Example:
 * <code>
 * &lt;binding name="effects" value="template:%message-code"/&gt;
 * </code>
 *
 */
public class TemplateBinding extends AbstractBinding
{
    /**
     * The root object against which the nested property name is evaluated.
     */
    private final IComponent root;
    /**
     * The expression that have variables, as a string.
     */
    private String expression;
    /**
     * cache the variable expression and it's compiled binding.
     */
    private Map variableBindings;
    /**
     * the BindingFactory for creating the inside variable IBinding.
     */
    private BindingFactory nestedBindingFactory;

    /**
     * constructor
     *
     * @param description
     * @param location
     * @param valueConverter
     * @param root                 The root object against which the nested property name is evaluated.
     * @param expression           plain string or message code in page's specification file.
     * @param nestedBindingFactory Internal used BindingFactory
     */
    public TemplateBinding(String description, Location location, ValueConverter valueConverter,
                           IComponent root, String expression, BindingFactory nestedBindingFactory)
    {
        super(description, valueConverter, location);
        this.root = root;
        this.expression = extractExpression(expression);
        this.nestedBindingFactory = nestedBindingFactory;
        buildingVariableBindings();
    }

    /**
     * build variable binding map.
     */
    private void buildingVariableBindings()
    {
        if (variableBindings == null) {
            variableBindings = new HashMap();
        }
        String least = expression;
        int start;
        while ((start = least.indexOf("${")) != -1) {
            least = least.substring(start);
            int length = least.indexOf("}");
            String innerExp = least.substring(2, length);
            IBinding expBinding = nestedBindingFactory.createBinding(root, getDescription(), innerExp, getLocation());
            variableBindings.put("${" + innerExp + "}", expBinding);
            least = least.substring(length + 1);
        }
    }

    /**
     * Extract expression if it is indicated as message code.
     *
     * @param expr
     * @return real evaluated expression.
     */
    private String extractExpression(String expr)
    {
        if (expr.startsWith("%")) {
            String key = expr.substring(1);
            return root.getMessages().getMessage(key);
        }
        return expr;
    }

    /**
     * Evaluate the value with inside variables.
     *
     * @return binding value
     */
    public Object getObject()
    {
        String value = expression;
        Iterator it = variableBindings.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            String key = (String) entry.getKey();
            IBinding exp = (IBinding) entry.getValue();
            String replacement = "";
            Object object = exp.getObject();
            if (object != null) {
                replacement = object.toString();
            }
            value = StringUtils.replace(value, key, replacement);
        }
        return value;
    }

    /**
     * Should be variant because of inside ognl expression.
     *
     * @return indicate it is variant.
     * @see org.apache.tapestry.IBinding#isInvariant()
     */
    public boolean isInvariant()
    {
        return false;
    }
}
