/*******************************************************************************
 * Copyright (c) 2005 Gabriel Handford.
 * All rights reserved.
 *
 * Contributors - Gabriel Handford
 ********************************************************************************/
package net.sf.tacos.services.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.tacos.services.CategoryInfo;
import net.sf.tacos.services.PageInfo;
import net.sf.tacos.services.SiteMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hivemind.Resource;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.SAXReader;

/**
 * This class reads the sitemap configuration, and provides access to
 * basic relationship information, such as page categories, bread crumbs, and
 * other information, that the application may need to ease navigation.
 *
 * @author Gabriel Handford
 */
public class SiteMapImpl implements SiteMap {

    private static final Log log = LogFactory.getLog(SiteMapImpl.class);

    private Resource resource;
    private Document document;

    private List cachedCategoryList = null;
    private Map cachedPageMap = new HashMap();
    private Map cachedCategoryInfo = new HashMap();
    
    private boolean compact;
    
    
    /**
     * Empty constructor.
     */
    public SiteMapImpl() { }
    
    /**
     * Set the sitemap resource (xml).
     * @param resource The resource (xml) file
     */
    public void setResource(Resource resource) {
        this.resource = resource;        
    }
    
    /**
     * Initialize site map from a url (xml document).
     * @throws DocumentException on error
     * @throws IllegalStateException If no resource has been set.
     */
    public void initialize() throws DocumentException {
        if (resource == null)
            throw new IllegalStateException(
                    "No resource set, call setResource(..) before initialize()");
        if (resource.getResourceURL() == null)
            throw new IllegalStateException(
                    "No valid resource URL could be resolved:" + resource);
        log.debug("Sitemap resource: " + resource.getResourceURL());
        SAXReader reader = new SAXReader();
        document = reader.read(resource.getResourceURL());
        
        Element main = (Element)document.selectSingleNode("/sitemap");
        
        String val = main.attributeValue("compact");
        if (val != null && !val.trim().equals("")) 
            compact = Boolean.valueOf(val).booleanValue();
    }

    /**
     * Check if page name is contained in the page element tree.
     * @param parentName The page to look for.
     * @param pageName The page to start at (moves bottom-up).
     * @return True if the parent contains the page in its tree.
     */
    public boolean contains(String parentName, String pageName) {
        PageInfo pageInfo = getPageInfo(pageName);
        return containsImpl(parentName, pageInfo);
    }

    /**
     * Check if page name is contained in the page element tree.
     * @param parentName The page to look for.
     * @param pageInfo The page to start at (moves bottom-up).
     * @return True if the parent contains the page in its tree.
     */
    private boolean containsImpl(String parentName, PageInfo pageInfo) {
        if (pageInfo == null) return false;
        PageInfo parent = pageInfo.getParent();
        if (parent == null) return false;
        else if (parent.getName().equals(parentName)) return true;
        else return containsImpl(parentName, parent);
    }
    
    /**
     * Get page information.
     * @param name The page name
     * @return The page info
     */
    public PageInfo getPageInfo(String name) {
        return getPageInfoCacheImpl(name, null);
    }

    /**
     * Get page information (from cache).
     * @param name The page name.
     * @param source The page name source (for cycle detetion, first time is null).
     * @return The page.
     */
    private PageInfo getPageInfoCacheImpl(String name, String source) {
        PageInfo pi = (PageInfo)cachedPageMap.get(name);
        if (pi == null) {
            pi = getPageInfoImpl(name, source);
            if (pi != null) cachedPageMap.put(name, pi);
        } else {
            //log.debug("Returning cached page info: " + name);
        }
        return pi;
    }

    /**
     * Get page information.
     * @param name The page name
     * @param source Original page name (for cycle detection)
     * @return The page info
     */
    private PageInfo getPageInfoImpl(String name, String source) {
        if (document == null) throw new IllegalStateException("No document set, call setResource and initialize.");
        
        // Cycle check
        if (source != null && name.equals(source))
            throw new IllegalStateException("Found a cycle, " + name + "->" + source);
        else if (source == null)
            source = name;
        
        Element node = findPageNode(name);
        if (node != null) {
            
            PageInfo page = parsePageNode(node);
            
            Node pageCategoryNode = document.selectSingleNode("/sitemap/category//page[@name='" + name + "']");
            if (pageCategoryNode != null) {
                parseNodeTree(page, pageCategoryNode);
                //Because it will have been re-parsed
                return (PageInfo)cachedPageMap.get(page.getName());
            }
            
            return page;
        }
        
        return null;
    }
    
    /**
     * Finds the specified page node in the sitemap, first trying to find
     * it by name and then path.
     * @param page
     * @return The page, or null if not found.
     */
    protected Element findPageNode(String page)
    {
        Element element = (Element)document.selectSingleNode("/sitemap/page[@name='" + page + "']");
        if (element == null && compact)
        {
            element = (Element)document.selectSingleNode("/sitemap/category//page[@name='" + page + "']");
        }
        return element;       
    }
    
    /**
     * Parses a particular page node, doesn't parse
     * children or parent of {@link PageInfo}.
     * 
     * @param node Node to parse.
     * @return
     */
    protected PageInfo parsePageNode(Element node)
    {
        if (node == null || !node.getName().equals("page"))
            throw new IllegalArgumentException("Node was null or not a page node.");
        
        String name = node.attributeValue("name");
        String desc = node.attributeValue("desc");
        String perm = node.attributeValue("permission");
        String listItem = node.attributeValue("navigable");
        //Whether or not to list this page if a child.
        boolean list = true;
        if (listItem != null && !listItem.trim().equals("")) 
            list = Boolean.valueOf(listItem).booleanValue();
        
        PageInfo page = new PageInfoImpl(name, desc, perm, new ArrayList(), list);
        return page;
    }
    
    /**
     * Iterates through the category tree represented 
     * by this page node and parses all children and parents.
     * 
     * @param page The page that caused the tree parse
     * @param pageNode
     */
    protected void parseNodeTree(PageInfo page, Node pageNode)
    {
        //First find top-level parent
        Node parent = pageNode;
        while (parent.getParent() != null 
                && parent.getParent().getName().equals("page")) {
            parent = parent.getParent();
        }
        
        log.debug("Parsing tree for page <" + page.getName() + ">");
        //We can now parse all children of parent
        parsePageNodes(parent, page);
    }
    
    /**
     * Parses all children of specified node and adds them
     * to the page cache.
     * 
     * @param parent
     * @param source Page that caused the original parse.
     * @return The parent node that was parsed.
     */
    protected PageInfo parsePageNodes(Node parent, PageInfo source)
    {
        PageInfo parentInfo = parsePageNode(findPageNode(((Element)parent).attributeValue("name")));
        cachedPageMap.put(parentInfo.getName(), parentInfo);
        
        log.debug("Parent page name <" + parent.valueOf("@name") + ">");
        Iterator it = parent.selectNodes("*").iterator();
        while (it.hasNext()) {
            Node child = (Node)it.next();
            PageInfo childInfo = parsePageNodes(child, source);
            
            childInfo.setParent(parentInfo);
            parentInfo.addChild(childInfo);
        }
        
        return parentInfo;
    }
    
    /**
     * Get category listing.
     * @return The category list
     */
    public synchronized List getCategories() {
        if (document == null) throw new IllegalStateException("No document set, call setResource and initialize.");
        if (cachedCategoryList == null) {
            List nodes = document.selectNodes("/sitemap/category[@name]");
            cachedCategoryList = new ArrayList(nodes.size());
            for(int i = 0, size = nodes.size(); i < size; i++) {
                String category = ((Node)nodes.get(i)).valueOf("@name");
                cachedCategoryList.add(category);
            }
        }
        return cachedCategoryList;
    }

    /**
     * Get category info for named category.
     * @param name Category name
     * @return The category info
     * @see SiteMap#getCategoryInfo(java.lang.String)
     */
    public CategoryInfo getCategoryInfo(String name) {
        CategoryInfo ci = (CategoryInfo)cachedCategoryInfo.get(name);
        if (ci != null) {
            return ci;
        }
        
        if (document == null) throw new IllegalStateException("No document set, call setResource and initialize.");

        String imgName = null;
        String inactiveImgName = null;
        
        Element catnode = (Element)document.selectSingleNode("/sitemap/category[@name='" + name + "']");
        if (catnode != null) {
            imgName = catnode.attributeValue("image-active");
            inactiveImgName = catnode.attributeValue("image-inactive");
        }
        
        //Node node = document.selectSingleNode("/sitemap/category[@name='" + name + "']");
        List nodes = document.selectNodes("/sitemap/category[@name='" + name + "']/page");
        if (nodes != null) {
            List pageNames = new ArrayList(nodes.size());
            for(int i = 0; i < nodes.size(); i++) {
                Node node = (Node)nodes.get(i);
                String pageName = node.valueOf("@name");
                pageNames.add(pageName);
            }
            ci = new CategoryInfoImpl(name, pageNames, imgName, inactiveImgName);
            cachedCategoryInfo.put(name, ci);
            return ci;
        }
        ci = new CategoryInfoImpl(name, null);
        cachedCategoryInfo.put(name, ci);
        return ci;
    }

    /**
     * Traverse up the tree to find the super parent.
     * @param pageInfo The page.
     * @return The topmost page.
     */
    private PageInfo getOrigin(PageInfo pageInfo) {
        if (pageInfo != null && pageInfo.getParent() != null) return getOrigin(pageInfo.getParent());
        return pageInfo;
    }

    /**
     * Get the category info for the page name.
     * @param pageName The page name.
     * @return The first category associated with this page.
     */
    public CategoryInfo getCategoryFromPage(String pageName) {
        PageInfo pageInfo = getPageInfo(pageName);
        if (pageInfo != null) {
            PageInfo topMostPage = getOrigin(pageInfo);
            String topPageName = topMostPage.getName();

            Node node = document.selectSingleNode("/sitemap/category/page[@name='" + topPageName + "']");
            if (node != null) {
                String categoryName = node.getParent().valueOf("@name");
                return getCategoryInfo(categoryName);
            }
        }
        return null;
    }
    
    /**
     * Check if page name is in the specified category.
     * @param pageName Page name
     * @param category Category
     * @return True if in the specified category
     */
    public boolean inCategory(String pageName, String category) {                
        CategoryInfo ci = getCategoryFromPage(pageName);
        return (ci != null && ci.getName().equals(category));
    }
    
    /**
     * Get the default page.
     * @param category Category
     * @return Default page
     */
    public PageInfo getDefaultPage(String category) {
        CategoryInfo ci = getCategoryInfo(category);
        if (ci == null) return null;
        return getPageInfo(ci.getDefaultPage());
    }
    
    /**
     * Get pages for a specific page names category.
     * @param pageName Page name
     * @return Pages
     */
    public List getCategoryPages(String pageName) {
        CategoryInfo ci = getCategoryFromPage(pageName);
        if (ci == null) return Arrays.asList(new Object[0]);
        return ci.getPageNames();        
    }
    
    /**
     * Get the default page description.
     * @param category Category
     * @return Category default page description
     */
    public String getDefaultPageDesc(String category) {
        PageInfo pi = getDefaultPage(category);
        if (pi == null) return null;
        return pi.getDesc();
    }

    /**
     * Get bread crumbs.
     * @param pageName The page name.
     * @return The bread crumbs (page name list).
     */
    public List getBreadCrumbs(String pageName) {
        List list = new LinkedList();
        PageInfo pageInfo = getPageInfo(pageName);
        if (pageInfo == null) return Arrays.asList(new Object[0]);
        loadBreadCrumbs(pageInfo, list);
        return list;
    }
    
    /**
     * Load bread crumbs.
     * @param pageInfo The page.
     * @param breadCrumbs The list.
     */
    private void loadBreadCrumbs(PageInfo pageInfo, List breadCrumbs) {
        PageInfo parent = pageInfo.getParent();
        if (parent == null || parent.getName().equals(""))
            return;

        breadCrumbs.add(0, parent.getName());
        loadBreadCrumbs(parent, breadCrumbs);
    }
    
    public boolean getCompact()
    {
        return compact;
    }

}
